/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL "http://www.cecill.info".                              # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include "sctk_spinlock.h"
#include "sctk_asm.h"
#include <sched.h>
#include <stdlib.h>
#include <math.h>

#define SCTK_SPIN_DELAY 10

volatile double sctk_dummy_results = 1.0;

static void
sctk_wait_func_dummy (int val)
{
  int i;
  double tmp;
  tmp = sctk_dummy_results;
  for (i = 0; i < val; i++)
    {
      sctk_dummy_results = pow (tmp, sin (tmp));
    }
  sctk_dummy_results = tmp;
}

void
sctk_spinlock_lock_no_yield (sctk_spinlock_t * lock)
{
  int val = 1;
  while (expect_true (sctk_test_and_set (lock)))
    {
      while (*lock != 0)
	{
	  val = val * 2;
	  if (val > 100)
	    {
	      val = 1;
	    }

#ifdef SCTK_ia64_ARCH_SCTK
	  sctk_hint_pause ();
#endif
	  sctk_wait_func_dummy (val);
	}
    }
}

#if 0
void
sctk_spinlock_lock (sctk_spinlock_t * lock)
{
  while (expect_true (sctk_test_and_set (lock)))
    {
      while (*lock != 0)
	{
#ifdef SCTK_ia64_ARCH_SCTK
	  sctk_hint_pause ();
#endif
	  sched_yield ();
	}
    }
}
#else
void
sctk_spinlock_lock (sctk_spinlock_t * lock)
{
  long i = SCTK_SPIN_DELAY;
  while (expect_true (sctk_test_and_set (lock)))
    {
      while (expect_true (*lock != 0))
	{
	  i--;
	  if (expect_false (i <= 0))
	    {
	      sched_yield ();
	      i = SCTK_SPIN_DELAY;
	    }
#ifdef SCTK_ia64_ARCH_SCTK
	  else
	    sctk_hint_pause ();
#endif
	}
    }
}
#endif

int
sctk_spinlock_trylock (sctk_spinlock_t * lock)
{
  return sctk_test_and_set (lock);
}

void
sctk_spinlock_unlock (sctk_spinlock_t * lock)
{
  *lock = 0;
}
