/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL "http://www.cecill.info".                              # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include <string.h>
#include <sched.h>
#include "sctk_inter_thread_comm.h"
#include "sctk_thread.h"
#include "sctk_alloc.h"
#include "sctk_asm.h"
#include "sctk_low_level_comm.h"
#include "sctk_accessor.h"

static sctk_ptp_data_t *sctk_ptp_list;
static volatile int *volatile sctk_ptp_process_localisation = NULL;
static sctk_alloc_thread_data_t *tls_inter_thread_comm = NULL;
static int sctk_total_task_number = 0;

#define SCTK_LOCAL_VERSION_MAJOR 0
#define SCTK_LOCAL_VERSION_MINOR 1

void
sctk_ptp_per_task_init_com (sctk_task_ptp_data_t * task, int k)
{
  task->source_task = k;

  task->send_list = NULL;
  task->send_list_tail = NULL;

  task->recv_list = NULL;
  task->recv_list_tail = NULL;

  task->is_usable = 1;
  sctk_thread_mutex_init (&(task->lock), NULL);
  task->rank_send = 0;
  task->rank_recv = 0;
}

void
sctk_ptp_per_task_init (int i)
{
  int j, k;
  int nb_task;
  sctk_alloc_thread_data_t *tls;

  sctk_ptp_list[i].tls = __sctk_create_thread_memory_area ();
  tls = sctk_ptp_list[i].tls;

  nb_task = sctk_total_task_number;

  for (j = 0; j < SCTK_MAX_COMMUNICATOR_NUMBER; j++)
    {
      sctk_ptp_list[i].communicators[j].wait_list = NULL;
      sctk_ptp_list[i].communicators[j].wait_list_tail = NULL;
      sctk_ptp_list[i].communicators[j].wait_send_list = NULL;
      sctk_ptp_list[i].communicators[j].wait_send_list_tail = NULL;
      sctk_ptp_list[i].communicators[j].tasks = (sctk_task_ptp_data_t *)
	__sctk_calloc (nb_task, sizeof (sctk_task_ptp_data_t), tls);
      assume (sctk_ptp_list[i].communicators[j].tasks != NULL);

      sctk_ptp_list[i].communicators[j].nb_tasks = nb_task;
      for (k = 0; k < nb_task; k++)
	{
	  sctk_ptp_per_task_init_com (&
				      (sctk_ptp_list[i].communicators[j].
				       tasks[k]), k);
	}
    }
  sctk_ptp_list[i].is_usable = 1;
}

void
sctk_ptp_init (const int nb_task)
{
  int i;
  sctk_only_once ();
  sctk_total_task_number = nb_task;
  sctk_print_version ("PtP", SCTK_LOCAL_VERSION_MAJOR,
		      SCTK_LOCAL_VERSION_MINOR);
  tls_inter_thread_comm = __sctk_create_thread_memory_area ();

  sctk_ptp_list =
    (sctk_ptp_data_t *) __sctk_calloc (nb_task, sizeof (sctk_ptp_data_t),
				       tls_inter_thread_comm);
  assume (sctk_ptp_list != NULL);

  sctk_ptp_process_localisation =
    (int *) __sctk_calloc (nb_task, sizeof (int), tls_inter_thread_comm);
  assume (sctk_ptp_process_localisation != NULL);

  for (i = 0; i < nb_task; i++)
    {
      sctk_ptp_process_localisation[i] = -1;
    }
  for (i = 0; i < nb_task; i++)
    {
      sctk_ptp_list[i].tls = NULL;
      sctk_ptp_list[i].free_list = NULL;
      sctk_ptp_list[i].is_usable = 0;
    }
}

void
sctk_ptp_delete ()
{
  int i, j;
  for (i = 0; i < sctk_total_task_number; i++)
    {
      for (j = 0; j < SCTK_MAX_COMMUNICATOR_NUMBER; j++)
	{
	  sctk_free ((void *) (sctk_ptp_list[i].communicators[j].tasks));
	}
    }
  sctk_free (sctk_ptp_list);
  sctk_free ((void *) sctk_ptp_process_localisation);
}

static inline void
sctk_copy_buffer_std_std (sctk_pack_indexes_t * restrict in_begins,
			  sctk_pack_indexes_t * restrict in_ends,
			  size_t in_sizes,
			  void *restrict in_adress,
			  size_t in_elem_size,
			  sctk_pack_indexes_t * restrict out_begins,
			  sctk_pack_indexes_t * restrict out_ends,
			  size_t out_sizes,
			  void *restrict out_adress, size_t out_elem_size)
{
  sctk_pack_indexes_t tmp_begin[1];
  sctk_pack_indexes_t tmp_end[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin;
	  in_begins[0] = 0;
	  in_ends = tmp_end;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin;
	  out_begins[0] = 0;
	  out_ends = tmp_end;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      memcpy (&(((char *) out_adress)[j]),
		      &(((char *) in_adress)[in_j]), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));

	      j += max_length;
	      in_j += max_length;

	    }
	}
    }
}

static inline void
sctk_copy_buffer_absolute_absolute (sctk_pack_absolute_indexes_t *
				    restrict in_begins,
				    sctk_pack_absolute_indexes_t *
				    restrict in_ends, size_t in_sizes,
				    void *restrict in_adress,
				    size_t in_elem_size,
				    sctk_pack_absolute_indexes_t *
				    restrict out_begins,
				    sctk_pack_absolute_indexes_t *
				    restrict out_ends, size_t out_sizes,
				    void *restrict out_adress,
				    size_t out_elem_size)
{
  sctk_pack_absolute_indexes_t tmp_begin[1];
  sctk_pack_absolute_indexes_t tmp_end[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin;
	  in_begins[0] = 0;
	  in_ends = tmp_end;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin;
	  out_begins[0] = 0;
	  out_ends = tmp_end;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      sctk_nodebug ("Copy out[%lu-%lu]%p == in[%lu-%lu]%p", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));
	      memcpy (&(((char *) out_adress)[j]),
		      &(((char *) in_adress)[in_j]), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));

	      j += max_length;
	      in_j += max_length;
	    }
	}
    }
}

static inline void
sctk_copy_buffer_absolute_std (sctk_pack_absolute_indexes_t *
			       restrict in_begins,
			       sctk_pack_absolute_indexes_t *
			       restrict in_ends, size_t in_sizes,
			       void *restrict in_adress,
			       size_t in_elem_size,
			       sctk_pack_indexes_t * restrict out_begins,
			       sctk_pack_indexes_t * restrict out_ends,
			       size_t out_sizes, void *restrict out_adress,
			       size_t out_elem_size)
{
  sctk_pack_indexes_t tmp_begin[1];
  sctk_pack_indexes_t tmp_end[1];
  sctk_pack_absolute_indexes_t tmp_begin_absolute[1];
  sctk_pack_absolute_indexes_t tmp_end_absolute[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin_absolute;
	  in_begins[0] = 0;
	  in_ends = tmp_end_absolute;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin;
	  out_begins[0] = 0;
	  out_ends = tmp_end;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      memcpy (&(((char *) out_adress)[j]),
		      &(((char *) in_adress)[in_j]), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));

	      j += max_length;
	      in_j += max_length;
	    }
	}
    }
}

static inline void
sctk_copy_buffer_std_absolute (sctk_pack_indexes_t * restrict in_begins,
			       sctk_pack_indexes_t * restrict in_ends,
			       size_t in_sizes,
			       void *restrict in_adress,
			       size_t in_elem_size,
			       sctk_pack_absolute_indexes_t *
			       restrict out_begins,
			       sctk_pack_absolute_indexes_t *
			       restrict out_ends, size_t out_sizes,
			       void *restrict out_adress,
			       size_t out_elem_size)
{
  sctk_pack_indexes_t tmp_begin[1];
  sctk_pack_indexes_t tmp_end[1];
  sctk_pack_absolute_indexes_t tmp_begin_absolute[1];
  sctk_pack_absolute_indexes_t tmp_end_absolute[1];
  if ((in_begins == NULL) && (out_begins == NULL))
    {
      sctk_nodebug ("%s == %s", out_adress, in_adress);
      memcpy (out_adress, in_adress, in_sizes);
      sctk_nodebug ("%s == %s", out_adress, in_adress);
    }
  else
    {
      unsigned long i;
      unsigned long j;
      unsigned long in_i;
      unsigned long in_j;
      if (in_begins == NULL)
	{
	  in_begins = tmp_begin;
	  in_begins[0] = 0;
	  in_ends = tmp_end;
	  in_ends[0] = in_sizes - 1;
	  in_elem_size = 1;
	  in_sizes = 1;
	}
      if (out_begins == NULL)
	{
	  out_begins = tmp_begin_absolute;
	  out_begins[0] = 0;
	  out_ends = tmp_end_absolute;
	  out_ends[0] = out_sizes - 1;
	  out_elem_size = 1;
	  out_sizes = 1;
	}
      in_i = 0;
      in_j = in_begins[in_i] * in_elem_size;
      for (i = 0; i < out_sizes; i++)
	{
	  for (j = out_begins[i] * out_elem_size;
	       j <= out_ends[i] * out_elem_size;)
	    {
	      size_t max_length;
	      if (in_j > in_ends[in_i] * in_elem_size)
		{
		  in_i++;
		  if (in_i > in_sizes)
		    {
		      return;
		    }
		  in_j = in_begins[in_i] * in_elem_size;
		}

	      max_length =
		sctk_min ((out_ends[i] * out_elem_size - j +
			   out_elem_size),
			  (in_ends[in_i] * in_elem_size - in_j +
			   in_elem_size));

	      memcpy (&(((char *) out_adress)[j]),
		      &(((char *) in_adress)[in_j]), max_length);
	      sctk_nodebug ("Copy out[%d-%d]%s == in[%d-%d]%s", j,
			    j + max_length, &(((char *) out_adress)[j]),
			    in_j, in_j + max_length,
			    &(((char *) in_adress)[in_j]));

	      j += max_length;
	      in_j += max_length;
	    }
	}
    }
}

static inline void
sctk_copy_buffer (sctk_pack_indexes_t * restrict in_begins,
		  sctk_pack_indexes_t * restrict in_ends,
		  sctk_pack_absolute_indexes_t *
		  restrict in_begins_absolute,
		  sctk_pack_absolute_indexes_t * restrict in_ends_absolute,
		  size_t in_sizes, void *restrict in_adress,
		  size_t in_elem_size,
		  sctk_pack_indexes_t * restrict out_begins,
		  sctk_pack_indexes_t * restrict out_ends,
		  sctk_pack_absolute_indexes_t *
		  restrict out_begins_absolute,
		  sctk_pack_absolute_indexes_t * restrict out_ends_absolute,
		  size_t out_sizes, void *restrict out_adress,
		  size_t out_elem_size)
{
  sctk_nodebug ("in %p out %p", in_adress, out_adress);
  if ((in_begins_absolute == NULL) && (out_begins_absolute == NULL))
    {
      sctk_copy_buffer_std_std (in_begins,
				in_ends,
				in_sizes,
				in_adress,
				in_elem_size,
				out_begins,
				out_ends,
				out_sizes, out_adress, out_elem_size);
    }
  else if ((in_begins_absolute != NULL) && (out_begins_absolute != NULL))
    {
      sctk_copy_buffer_absolute_absolute (in_begins_absolute,
					  in_ends_absolute, in_sizes,
					  in_adress, in_elem_size,
					  out_begins_absolute,
					  out_ends_absolute, out_sizes,
					  out_adress, out_elem_size);
    }
  else if ((in_begins_absolute == NULL) && (out_begins_absolute != NULL))
    {
      sctk_copy_buffer_std_absolute (in_begins, in_ends, in_sizes,
				     in_adress, in_elem_size,
				     out_begins_absolute,
				     out_ends_absolute, out_sizes,
				     out_adress, out_elem_size);
    }
  else
    {
      sctk_copy_buffer_absolute_std (in_begins_absolute,
				     in_ends_absolute,
				     in_sizes,
				     in_adress,
				     in_elem_size,
				     out_begins,
				     out_ends,
				     out_sizes, out_adress, out_elem_size);
    }
}

static inline void
sctk_copy_message_net (sctk_thread_ptp_message_t * restrict dest,
		       sctk_thread_ptp_message_t * restrict src)
{
  int i;
  char *cursor;
  sctk_nodebug ("Copy a message");
  cursor = src->net_mesg;

  if (cursor != (char *) (-1))
    {
      for (i = 0; i < dest->message.nb_items; i++)
	{
	  if (dest->message.begins_absolute[i] != NULL)
	    {
	      int j;
	      char *tmp;
	      size_t offset;
	      size_t size;
	      sctk_nodebug ("Copy a message pack");
	      for (j = 0; j < dest->message.sizes[i]; j++)
		{
		  tmp = dest->message.adresses[i];
		  offset = dest->message.begins_absolute[i][j] *
		    dest->message.elem_sizes[i];

		  size =
		    (dest->message.ends_absolute[i][j] -
		     dest->message.begins_absolute[i][j] + 1) *
		    (dest->message.elem_sizes[i]);

		  memcpy (&(tmp[offset]), cursor, size);
		  cursor += size;
		}
	    }
	  else
	    {
	      if (dest->message.begins[i] == NULL)
		{
		  sctk_nodebug ("Copy a message no pack");
		  sctk_nodebug ("%s == %s", dest->message.adresses[i],
				cursor);
		  memcpy (dest->message.adresses[i], cursor,
			  dest->message.sizes[i]);
		  cursor += dest->message.sizes[i];
		}
	      else
		{
		  int j;
		  char *tmp;
		  size_t offset;
		  size_t size;
		  sctk_nodebug ("Copy a message pack");
		  for (j = 0; j < dest->message.sizes[i]; j++)
		    {
		      tmp = dest->message.adresses[i];
		      offset = dest->message.begins[i][j] *
			dest->message.elem_sizes[i];

		      size =
			(dest->message.ends[i][j] -
			 dest->message.begins[i][j] + 1) *
			(dest->message.elem_sizes[i]);

		      memcpy (&(tmp[offset]), cursor, size);
		      cursor += size;
		    }
		}
	    }
	}
      sctk_nodebug ("Free %p", src);
      sctk_net_free_func ((sctk_thread_ptp_message_t *) src);
    }
  else
    {
      sctk_net_copy_message_func (dest, src);
    }
}

static inline sctk_thread_ptp_message_t *
sctk_find_matching (const
		    sctk_thread_ptp_message_t
		    * restrict dest, sctk_task_ptp_data_t * restrict data)
{
  register sctk_thread_ptp_message_t *restrict cursor;
  sctk_thread_ptp_message_t *restrict tmp;
  sctk_thread_ptp_message_t *restrict last;
  register int dest_tag;

  dest_tag = dest->header.message_tag;

  cursor = (sctk_thread_ptp_message_t *) data->send_list;

  if (cursor != NULL)
    {
      tmp = cursor->next;
      if (dest_tag == cursor->header.message_tag)
	{
	  cursor->next = NULL;
	  if (tmp == NULL)
	    {
	      data->send_list_tail = NULL;
	    }
	  data->send_list = tmp;
	  return cursor;
	}
      last = cursor;
      cursor = tmp;
    }
  else
    {
      return NULL;
    }

  while (cursor != NULL)
    {
      tmp = cursor->next;
      if (dest_tag == cursor->header.message_tag)
	{
	  cursor->next = NULL;
	  last->next = tmp;
	  if (tmp == NULL)
	    {
	      data->send_list_tail = last;
	    }
	  return cursor;
	}
      last = cursor;
      cursor = tmp;
    }
  return NULL;
}

static inline void
__sctk_memcpy (char *restrict s1, const char *restrict s2, size_t n)
{
  if (s1 != s2)
    {
      memcpy (s1, s2, n);
    }
}

static inline int
__sctk_perform_match_for_source (sctk_task_ptp_data_t * restrict data)
{
  int res = 0;
  sctk_thread_ptp_message_t *restrict cursor;
  sctk_thread_ptp_message_t *restrict src;
  sctk_thread_ptp_message_t *restrict new_list = NULL;
  sctk_thread_ptp_message_t *restrict new_list_tail = NULL;

  if ((data->recv_list != NULL) && (data->send_list != NULL))
    {
      while (sctk_thread_mutex_trylock (&data->lock) != 0)
	sched_yield ();
      cursor = (sctk_thread_ptp_message_t *) data->recv_list;
      while (cursor != NULL)
	{
	  sctk_thread_ptp_message_t *restrict tmp;
	  tmp = cursor->next;
	  cursor->next = NULL;
	  src = sctk_find_matching (cursor, data);
	  if (src != NULL)
	    {
	      cursor->header.msg_size = src->header.msg_size;

	      /*Message match */
	      if (src->net_mesg == NULL)
		{
		  sctk_nodebug ("There are %d parts",
				cursor->message.nb_items);
		  if ((cursor->message.nb_items == 1)
		      && (src->message.begins[0] == NULL)
		      && (cursor->message.begins[0] == NULL)
		      && (src->message.begins_absolute[0] == NULL)
		      && (cursor->message.begins_absolute[0] == NULL))
		    {
		      __sctk_memcpy (cursor->message.adresses[0],
				     src->message.adresses[0],
				     src->message.sizes[0]);
		    }
		  else
		    {
		      int i;
		      for (i = 0; i < cursor->message.nb_items; i++)
			{
			  sctk_copy_buffer (src->message.begins[i],
					    src->message.ends[i],
					    src->message.
					    begins_absolute[i],
					    src->message.ends_absolute[i],
					    src->message.sizes[i],
					    src->message.adresses[i],
					    src->message.elem_sizes[i],
					    cursor->message.begins[i],
					    cursor->message.ends[i],
					    cursor->message.
					    begins_absolute[i],
					    cursor->message.
					    ends_absolute[i],
					    cursor->message.sizes[i],
					    cursor->message.adresses[i],
					    cursor->message.elem_sizes[i]);
			}
		    }
		  sctk_nodebug ("Wake %p", &(src->completion_flag));
		  src->completion_flag = 1;
		}
	      else
		{
		  sctk_copy_message_net (cursor, src);
		}

	      /*Just flip the completion flag, memory liberation 
	         will be performed by the caller */
	      sctk_nodebug ("Wake %p", &(cursor->completion_flag));
	      cursor->completion_flag = 1;
	    }
	  else
	    {
	      res++;
	      /*Message not ready */
	      if (new_list == NULL)
		{
		  new_list = cursor;
		}
	      else
		{
		  new_list_tail->next = cursor;
		}
	      new_list_tail = cursor;
	    }
	  cursor = tmp;
	}
      data->recv_list = new_list;
      data->recv_list_tail = new_list_tail;
      sctk_thread_mutex_unlock (&data->lock);
    }
  return res;
}

static inline int
__sctk_perform_match_any_source_wait (sctk_per_communicator_ptp_data_t *
				      communicator)
{
  sctk_thread_ptp_message_t *cursor;
  int res = 0;
  cursor = communicator->wait_list;
  while (cursor != NULL)
    {
      if ((cursor->completion_flag == 0) &&
	  (cursor->data->recv_list != NULL) &&
	  (cursor->data->send_list != NULL))
	{
	  __sctk_perform_match_for_source (cursor->data);
	}
      if (cursor->completion_flag == 0)
	{
	  res++;
	}
      cursor = cursor->wait_next;
    }
  return res;
}
static inline void
__sctk_free_message_wait (sctk_per_communicator_ptp_data_t *
			  restrict communicator,
			  sctk_ptp_data_t * restrict from)
{
  sctk_thread_ptp_message_t *restrict cursor;
  sctk_thread_ptp_message_t *restrict new_list = NULL;
  sctk_thread_ptp_message_t *restrict new_list_tail = NULL;
  sctk_thread_ptp_message_t *restrict new_free_list = NULL;
  sctk_thread_ptp_message_t *restrict new_free_list_tail = NULL;
  cursor = communicator->wait_list;
  while (cursor != NULL)
    {
      sctk_thread_ptp_message_t *restrict tmp = NULL;
      tmp = cursor->wait_next;
      cursor->wait_next = NULL;
      if (cursor->completion_flag == 0)
	{
	  if (new_list == NULL)
	    {
	      new_list = cursor;
	      new_list_tail = cursor;
	    }
	  else
	    {
	      new_list_tail->wait_next = cursor;
	      new_list_tail = cursor;
	    }
	}
      else
	{
	  sctk_nodebug ("free %p", cursor);
	  if (cursor->header.request != NULL)
	    {
	      *(cursor->header.request) = cursor->completion_flag;
	      *(cursor->header.req_msg_size) = cursor->header.msg_size;
	    }
	  if (cursor->tls != NULL)
	    {
	      if (new_free_list == NULL)
		{
		  new_free_list = cursor;
		  new_free_list_tail = cursor;
		}
	      else
		{
		  new_free_list_tail->wait_next = cursor;
		  new_free_list_tail = cursor;
		}
	    }
	}
      cursor = tmp;
    }
  communicator->wait_list = new_list;
  communicator->wait_list_tail = new_list_tail;
  if (new_free_list_tail != NULL)
    {
      new_free_list_tail->wait_next = from->free_list;
      from->free_list = new_free_list;
    }

  new_list = NULL;
  new_list_tail = NULL;
  new_free_list = NULL;
  new_free_list_tail = NULL;
  cursor = communicator->wait_send_list;
  while (cursor != NULL)
    {
      sctk_thread_ptp_message_t *restrict tmp = NULL;
      tmp = cursor->wait_next;
      cursor->wait_next = NULL;
      if (cursor->completion_flag == 0)
	{
	  if (new_list == NULL)
	    {
	      new_list = cursor;
	      new_list_tail = cursor;
	    }
	  else
	    {
	      new_list_tail->wait_next = cursor;
	      new_list_tail = cursor;
	    }
	}
      else
	{
	  sctk_nodebug ("free %p", cursor);
	  if (cursor->header.request != NULL)
	    {
	      *(cursor->header.request) = cursor->completion_flag;
	      *(cursor->header.req_msg_size) = cursor->header.msg_size;
	    }
	  if (cursor->tls != NULL)
	    {
	      if (new_free_list == NULL)
		{
		  new_free_list = cursor;
		  new_free_list_tail = cursor;
		}
	      else
		{
		  new_free_list_tail->wait_next = cursor;
		  new_free_list_tail = cursor;
		}
	    }
	}
      cursor = tmp;
    }
  communicator->wait_send_list = new_list;
  communicator->wait_send_list_tail = new_list_tail;
  if (new_free_list_tail != NULL)
    {
      new_free_list_tail->wait_next = from->free_list;
      from->free_list = new_free_list;
    }
}

static inline void
__sctk_probe_for_source (sctk_task_ptp_data_t * data,
			 sctk_thread_message_header_t * msg, int *state)
{
  *state = 0;
  if (data->send_list != NULL)
    {
      *state = 1;
      *msg = data->send_list->header;
    }
}

static inline void
__sctk_probe_for_source_tag (sctk_task_ptp_data_t * data,
			     sctk_thread_message_header_t * msg, int *state)
{
  int message_tag;
  volatile sctk_thread_ptp_message_t *tmp;
  sctk_thread_mutex_spinlock (&data->lock);
  message_tag = msg->message_tag;
  *state = 0;
  tmp = data->send_list;
  while (tmp != NULL)
    {
      if (tmp->header.message_tag == message_tag)
	{
	  *state = 1;
	  *msg = tmp->header;
	  break;
	}
      tmp = tmp->next;
    }
  sctk_thread_mutex_unlock (&data->lock);
}

static inline void
__sctk_probe_any_source (sctk_per_communicator_ptp_data_t * communicator,
			 sctk_thread_message_header_t * msg, int *state)
{
  int i;
  *state = 0;
  for (i = 0; i < communicator->nb_tasks; i++)
    {
      __sctk_probe_for_source (&(communicator->tasks[i]), msg, state);
      if (*state == 1)
	{
	  return;
	}
    }
}

static inline int
sctk_place_message_send (sctk_ptp_data_t * dest,
			 sctk_thread_ptp_message_t * msg)
{
  sctk_task_ptp_data_t *data;
  int res = 0;

  sctk_nodebug
    ("TRY TO PLACE in comm %d in task %d with tag %d form %d is usable %d",
     msg->header.communicator, msg->header.destination,
     msg->header.message_tag, msg->header.source, dest->is_usable);
  if (dest->is_usable == 0)
    return 1;

  if (msg->header.communicator == -1 ||
      msg->header.destination == -1 ||
      msg->header.message_tag == -1 || msg->header.source == -1)
    {
      sctk_nodebug ("PLACE in comm %d in task %d with tag %d form %d",
		    msg->header.communicator, msg->header.destination,
		    msg->header.message_tag, msg->header.source);
    }

  data =
    &(dest->communicators[msg->header.communicator].
      tasks[msg->header.source]);
  msg->data = NULL;

/*   while (sctk_thread_mutex_trylock (&data->lock) != 0) sched_yield(); */
  sctk_thread_mutex_lock (&data->lock);
  if (data->is_usable)
    {
      msg->data = data;

      msg->header.rank = data->rank_send;
      data->rank_send++;

      if (data->send_list != NULL)
	{
	  data->send_list_tail->next = msg;
	}
      else
	{
	  data->send_list = msg;
	}
      data->send_list_tail = msg;
    }
  else
    {
      res = 1;
    }
  sctk_thread_mutex_unlock (&data->lock);

  sctk_nodebug ("Message send in %p %d return %d", data, data->is_usable,
		res);

  return res;
}

void
sctk_check_message_match (sctk_thread_ptp_message_t * msg)
{
  sctk_ptp_data_t *dest;
  sctk_task_ptp_data_t *data;

  dest = &(sctk_ptp_list[msg->header.destination]);

  if (dest->is_usable == 0)
    return;

  data =
    &(dest->communicators[msg->header.communicator].
      tasks[msg->header.source]);
  __sctk_perform_match_for_source (data);
}

void
sctk_check_messages (int destination, int source,
		     sctk_communicator_t communicator)
{
  sctk_ptp_data_t *dest;
  sctk_task_ptp_data_t *data;

  dest = &(sctk_ptp_list[destination]);

  if (dest->is_usable == 0)
    return;

  data = &(dest->communicators[communicator].tasks[source]);
  __sctk_perform_match_for_source (data);
}

static inline int
sctk_place_message_recv (sctk_ptp_data_t * dest,
			 sctk_thread_ptp_message_t * msg)
{
  sctk_task_ptp_data_t *data;
/*   assume(dest->is_usable != 0); */

  data =
    &(dest->communicators[msg->header.communicator].
      tasks[msg->header.source]);

  msg->header.rank = data->rank_recv;
  data->rank_recv++;

  sctk_nodebug ("Message recv in %p", data);
  msg->data = data;
  if (data->recv_list != NULL)
    {
      data->recv_list_tail->next = msg;
    }
  else
    {
      data->recv_list = msg;
    }
  data->recv_list_tail = msg;
  return 0;
}

static inline sctk_thread_ptp_message_t *
__sctk_message_alloc (int old_size, sctk_alloc_thread_data_t * tls)
{
  static int max_size = 0;
  int size;
  sctk_thread_ptp_message_t *tmp;
  char *ptr;

  size = old_size;

  if (max_size < old_size)
    {
      max_size = old_size;
    }
  if (max_size > old_size)
    {
      old_size = max_size;
    }

  size = sctk_max (old_size, size);

  size = ((size / SCTK_MESSAGE_UNIT_NUMBER) + 1) * SCTK_MESSAGE_UNIT_NUMBER;
  tmp = (sctk_thread_ptp_message_t *)
    __sctk_calloc (SCTK_MESSAGE_UNIT_SIZE (size), 1, tls);
  ptr = (char *) tmp;
  ptr += sctk_aligned_sizeof (sctk_thread_ptp_message_t);

  tmp->message.begins = (sctk_pack_indexes_t **) ptr;
  ptr += sctk_aligned_size (size * sizeof (sctk_pack_indexes_t *));
  tmp->message.ends = (sctk_pack_indexes_t **) ptr;
  ptr += sctk_aligned_size (size * sizeof (sctk_pack_indexes_t *));

  tmp->message.begins_absolute = (sctk_pack_absolute_indexes_t **) ptr;
  ptr += sctk_aligned_size (size * sizeof (sctk_pack_absolute_indexes_t *));
  tmp->message.ends_absolute = (sctk_pack_absolute_indexes_t **) ptr;
  ptr += sctk_aligned_size (size * sizeof (sctk_pack_absolute_indexes_t *));

  tmp->message.sizes = (size_t *) ptr;
  ptr += sctk_aligned_size (size * sizeof (size_t));
  tmp->message.elem_sizes = (size_t *) ptr;
  ptr += sctk_aligned_size (size * sizeof (size_t));
  tmp->message.adresses = (void **) ptr;

  tmp->message.nb_items_max = size;
  tmp->message.max_count = 0;
  tmp->message.nb_items = 0;
  tmp->message.message_size = 0;

  return tmp;
}


static inline sctk_thread_ptp_message_t *
__sctk_message_realloc (int new_size, sctk_thread_ptp_message_t * orig)
{
  sctk_nodebug ("Size %d, cur size %d", new_size, orig->message.nb_items_max);
  if (orig->message.nb_items_max > new_size)
    {
      return orig;
    }
  else
    {
      sctk_thread_ptp_message_t *tmp;
      int size;
      size = orig->message.nb_items_max;

      tmp = __sctk_message_alloc (new_size, orig->tls);

      memcpy (tmp->message.begins, orig->message.begins,
	      sctk_aligned_size (size * sizeof (sctk_pack_indexes_t *)));
      memcpy (tmp->message.ends, orig->message.ends,
	      sctk_aligned_size (size * sizeof (sctk_pack_indexes_t *)));

      memcpy (tmp->message.begins_absolute, orig->message.begins_absolute,
	      sctk_aligned_size (size *
				 sizeof (sctk_pack_absolute_indexes_t *)));
      memcpy (tmp->message.ends_absolute, orig->message.ends_absolute,
	      sctk_aligned_size (size *
				 sizeof (sctk_pack_absolute_indexes_t *)));

      memcpy (tmp->message.sizes, orig->message.sizes,
	      sctk_aligned_size (size * sizeof (size_t)));
      memcpy (tmp->message.elem_sizes, orig->message.elem_sizes,
	      sctk_aligned_size (size * sizeof (size_t)));
      memcpy (tmp->message.adresses, orig->message.adresses,
	      sctk_aligned_size (size * sizeof (void *)));

      tmp->header = orig->header;
      tmp->tls = orig->tls;

      tmp->message.nb_items = orig->message.nb_items;
      tmp->message.message_size = orig->message.message_size;
      tmp->message.max_count = orig->message.max_count;

      tmp->header.myself = orig->header.myself;
      sctk_free (orig);
      return tmp;
    }
}



sctk_thread_ptp_message_t *
sctk_create_header (const int myself)
{
  sctk_thread_ptp_message_t *restrict tmp;
  sctk_alloc_thread_data_t *restrict tls;
  int rank_comm_world;

  rank_comm_world = sctk_get_rank (SCTK_COMM_WORLD, sctk_get_task_rank ());

  sctk_nodebug ("Myself create %d glob %d ", myself, rank_comm_world);

  if (sctk_ptp_list[rank_comm_world].free_list != NULL)
    {
      tmp = sctk_ptp_list[rank_comm_world].free_list;
      sctk_ptp_list[rank_comm_world].free_list = tmp->wait_next;
      sctk_init_thread_ptp_message (tmp, myself);
      return tmp;
    }

  tls = sctk_ptp_list[rank_comm_world].tls;
  sctk_nodebug ("Myself create %d tls %p glob %d", myself, tls,
		rank_comm_world);

  tmp = __sctk_message_alloc (1, tls);
  tmp->tls = tls;
  sctk_init_thread_ptp_message (tmp, myself);
  return tmp;
}

sctk_thread_ptp_message_t
  * sctk_add_adress_in_message (sctk_thread_ptp_message_t * restrict msg,
				void *restrict adr, const size_t size)
{
  register int nb_items;
  nb_items = msg->message.nb_items;
  msg = __sctk_message_realloc (nb_items + 1, msg);

  msg->message.begins[nb_items] = NULL;
  msg->message.ends[nb_items] = NULL;

  msg->message.begins_absolute[nb_items] = NULL;
  msg->message.ends_absolute[nb_items] = NULL;

  msg->message.sizes[nb_items] = size;
  msg->message.elem_sizes[nb_items] = 1;
  msg->message.adresses[nb_items] = adr;

  msg->message.nb_items = nb_items + 1;
  if (msg->message.max_count < 1)
    {
      msg->message.max_count = 1;
    }
  return msg;
}

sctk_thread_ptp_message_t
  * sctk_add_pack_in_message (sctk_thread_ptp_message_t * msg, void *adr,
			      const sctk_count_t nb_items,
			      const size_t elem_size,
			      sctk_pack_indexes_t * begins,
			      sctk_pack_indexes_t * ends)
{
  register int nb_items_in_msg;
  nb_items_in_msg = msg->message.nb_items;
  msg = __sctk_message_realloc (nb_items_in_msg + 1, msg);

  msg->message.begins[nb_items_in_msg] = begins;
  msg->message.ends[nb_items_in_msg] = ends;

  msg->message.begins_absolute[nb_items_in_msg] = NULL;
  msg->message.ends_absolute[nb_items_in_msg] = NULL;

  msg->message.sizes[nb_items_in_msg] = nb_items;
  msg->message.elem_sizes[nb_items_in_msg] = elem_size;
  msg->message.adresses[nb_items_in_msg] = adr;

  msg->message.nb_items = nb_items_in_msg + 1;
  if (msg->message.max_count < nb_items)
    {
      msg->message.max_count = nb_items;
    }
  sctk_nodebug ("msg->message.max_count = %d", msg->message.max_count);
  return msg;
}

sctk_thread_ptp_message_t
  * sctk_add_pack_in_message_absolute (sctk_thread_ptp_message_t * msg,
				       void *adr,
				       const sctk_count_t nb_items,
				       const size_t elem_size,
				       sctk_pack_absolute_indexes_t *
				       begins,
				       sctk_pack_absolute_indexes_t * ends)
{
  register int nb_items_in_msg;
  nb_items_in_msg = msg->message.nb_items;
  msg = __sctk_message_realloc (nb_items_in_msg + 1, msg);

  msg->message.begins[nb_items_in_msg] = NULL;
  msg->message.ends[nb_items_in_msg] = NULL;

  msg->message.begins_absolute[nb_items_in_msg] = begins;
  msg->message.ends_absolute[nb_items_in_msg] = ends;

  msg->message.sizes[nb_items_in_msg] = nb_items;
  msg->message.elem_sizes[nb_items_in_msg] = elem_size;
  msg->message.adresses[nb_items_in_msg] = adr;

  msg->message.nb_items = nb_items_in_msg + 1;
  if (msg->message.max_count < nb_items)
    {
      msg->message.max_count = nb_items;
    }
  sctk_nodebug ("msg->message.max_count = %d", msg->message.max_count);
  return msg;
}

typedef struct
{
  int val;
  int task;
} sctk_placement_status_t;

static void
sctk_placement_test (sctk_placement_status_t * arg)
{
  if (sctk_ptp_process_localisation[arg->task] != -1)
    arg->val = 1;

}

int
sctk_is_net_message (int dest)
{
  sctk_ptp_data_t *tmp;
  int pos;
  tmp = &(sctk_ptp_list[dest]);
  if (tmp->is_usable == 0)
    return 0;

  pos = sctk_ptp_process_localisation[dest];

  if ((pos != sctk_process_rank) && (pos != -1))
    {
      return 1;
    }
  else
    {
      return 0;
    }
}

void
sctk_send_message (sctk_thread_ptp_message_t * msg)
{
  sctk_ptp_data_t *tmp;
  sctk_per_communicator_ptp_data_t *communicator;
  int is_net = 0;
  sctk_placement_status_t state;

  msg->wait_next = NULL;
  msg->next = NULL;

  sctk_nodebug ("send message %p to %d->%d", msg,
		msg->header.source, msg->header.destination);
  if (msg->net_mesg != NULL)
    {
      is_net = 1;
    }

  tmp = &(sctk_ptp_list[msg->header.destination]);
  while (sctk_place_message_send (tmp, msg) == 1)
    {
      int pos;
      if (sctk_ptp_process_localisation[msg->header.destination] == -1)
	{
	  state.val = 0;
	  state.task = msg->header.destination;
	  sctk_thread_wait_for_value_and_poll (&(state.val), 1,
					       (void (*)(void *))
					       sctk_placement_test, &state);
	}
      pos = sctk_ptp_process_localisation[msg->header.destination];
      if ((pos != sctk_process_rank) && (pos != -1))
	{
	  sctk_nodebug ("Task %d on %d", msg->header.destination, pos);
	  sctk_net_send_ptp_message (msg, pos);
	  break;
	}
      sctk_thread_yield ();
    }

  sctk_nodebug ("Task %d is here", msg->header.destination);

  if (is_net == 0)
    {
      /*Network messages are threated during copy */
      sctk_nodebug ("insert %p", msg);
      tmp = &(sctk_ptp_list[msg->header.myself]);
      communicator = &(tmp->communicators[msg->header.communicator]);
      if (communicator->wait_send_list != NULL)
	{
	  communicator->wait_send_list_tail->wait_next = msg;
	}
      else
	{
	  communicator->wait_send_list = msg;
	}
      communicator->wait_send_list_tail = msg;
    }
  sctk_nodebug ("send message %p done", msg);
}

void
sctk_recv_message (sctk_thread_ptp_message_t * msg)
{
  sctk_ptp_data_t *tmp;
  sctk_per_communicator_ptp_data_t *communicator;

  msg->wait_next = NULL;
  msg->next = NULL;
  sctk_nodebug ("recv message %p to %d->%d", msg,
		msg->header.source, msg->header.destination);

  tmp = &(sctk_ptp_list[msg->header.destination]);
  sctk_place_message_recv (tmp, msg);

  tmp = &(sctk_ptp_list[msg->header.myself]);
  communicator = &(tmp->communicators[msg->header.communicator]);
  if (communicator->wait_list != NULL)
    {
      communicator->wait_list_tail->wait_next = msg;
    }
  else
    {
      communicator->wait_list = msg;
    }
  communicator->wait_list_tail = msg;

  __sctk_perform_match_any_source_wait (communicator);
  __sctk_free_message_wait (communicator, tmp);
}

void
sctk_probe_source_any_tag (int destination, int source,
			   const sctk_communicator_t comm, int *status,
			   sctk_thread_message_header_t * msg)
{
  sctk_ptp_data_t *tmp;
  sctk_task_ptp_data_t *data;
/*   sctk_per_communicator_ptp_data_t * communicator; */

  tmp = &(sctk_ptp_list[sctk_translate_to_global_rank (destination, comm)]);
  data =
    &(tmp->communicators[comm].
      tasks[sctk_translate_to_global_rank (source, comm)]);
/*   communicator = &(tmp->communicators[comm]); */

  __sctk_probe_for_source (data, msg, status);
}

void
sctk_probe_source_tag (int destination, int source,
		       const sctk_communicator_t comm, int *status,
		       sctk_thread_message_header_t * msg)
{
  sctk_ptp_data_t *tmp;
  sctk_task_ptp_data_t *data;
/*   sctk_per_communicator_ptp_data_t * communicator; */

  tmp = &(sctk_ptp_list[sctk_translate_to_global_rank (destination, comm)]);
  data =
    &(tmp->communicators[comm].
      tasks[sctk_translate_to_global_rank (source, comm)]);
/*   communicator = &(tmp->communicators[comm]); */

  __sctk_probe_for_source_tag (data, msg, status);
}

void
sctk_probe_any_source_tag (int destination, int source,
			   const sctk_communicator_t comm, int *status,
			   sctk_thread_message_header_t * msg)
{
  sctk_ptp_data_t *tmp;
  sctk_task_ptp_data_t *data;
  int i;
  sctk_per_communicator_ptp_data_t *communicator;

  tmp = &(sctk_ptp_list[sctk_translate_to_global_rank (destination, comm)]);
  communicator = &(tmp->communicators[comm]);

  for (i = 0; i < communicator->nb_tasks; i++)
    {
      data = &(communicator->tasks[sctk_translate_to_global_rank (i, comm)]);
      __sctk_probe_for_source_tag (data, msg, status);
      if (*status == 1)
	{
	  return;
	}
    }
}

void
sctk_probe_any_source_any_tag (int destination,
			       const sctk_communicator_t comm, int *status,
			       sctk_thread_message_header_t * msg)
{
  sctk_ptp_data_t *tmp;
  sctk_per_communicator_ptp_data_t *communicator;

  tmp = &(sctk_ptp_list[sctk_translate_to_global_rank (destination, comm)]);
  communicator = &(tmp->communicators[comm]);

  __sctk_probe_any_source (communicator, msg, status);
}

void
sctk_register_thread (const int i)
{
  int j, k;
/*   sctk_ptp_data_t* tmp; */
/*   tmp = &(sctk_ptp_list[i]); */
  sctk_ptp_process_localisation[i] = sctk_process_rank;
  sctk_nodebug ("task %d on %d", i, sctk_ptp_process_localisation[i]);
  for (j = 0; j < SCTK_MAX_COMMUNICATOR_NUMBER; j++)
    {
      for (k = 0; k < sctk_ptp_list[i].communicators[j].nb_tasks; k++)
	{
	  sctk_ptp_list[i].communicators[j].tasks[k].is_usable = 1;
	}
    }
  sctk_net_send_task_location (i, sctk_process_rank);
}

void
sctk_register_distant_thread (const int i, const int pos)
{
  sctk_nodebug ("set task %d on %d really done", i, pos);
  while (sctk_ptp_process_localisation == NULL)
    sctk_thread_yield ();
  sctk_ptp_process_localisation[i] = pos;
}

void
sctk_register_restart_thread (const int i, const int pos)
{
  sctk_ptp_process_localisation[i] = pos;
  sctk_nodebug ("task %d on %d", i, sctk_ptp_process_localisation[i]);
}

void
sctk_unregister_distant_thread (const int i)
{
  sctk_ptp_process_localisation[i] = -1;
}

void
sctk_unregister_thread (const int i)
{
  int j, k;
/*   sctk_ptp_data_t* tmp; */
/*   tmp = &(sctk_ptp_list[i]); */
  sctk_nodebug ("unregister task %d on %d", i,
		sctk_ptp_process_localisation[i]);
  sctk_ptp_process_localisation[i] = -1;
  for (j = 0; j < SCTK_MAX_COMMUNICATOR_NUMBER; j++)
    {
      for (k = 0; k < sctk_ptp_list[i].communicators[j].nb_tasks; k++)
	{
	  sctk_thread_mutex_lock (&
				  (sctk_ptp_list[i].communicators[j].
				   tasks[k].lock));
	  sctk_ptp_list[i].communicators[j].tasks[k].is_usable = 0;
	  if (sctk_ptp_list[i].communicators[j].tasks[k].send_list != NULL)
	    {
	      /*clean dues to migration */
	      sctk_error
		("There's some pending messages from %d to %d on communicator %d",
		 i, j, k);
/* 	    sctk_error("Migration part of ptp communication not implemented"); */
	      not_reachable ();
	    }
	  sctk_thread_mutex_unlock (&
				    (sctk_ptp_list[i].communicators[j].
				     tasks[k].lock));
	}
    }
}

/* void sctk_check_for_communicator(const int task,const sctk_communicator_t comm){ */
/*   __sctk_perform_match_any_source_wait (&(sctk_ptp_list[task].communicators[comm])); */
/* } */

typedef struct
{
  int myself;
  sctk_communicator_t communicator;
  int *res;
} sctk_message_wait_t;

static void
sctk_check_for_communicator_poll (sctk_message_wait_t * restrict arg)
{
  int i;
  sctk_ptp_data_t *restrict tmp;
  sctk_per_communicator_ptp_data_t *restrict communicator;
  int val = 0;
  int res;

/*   if (sctk_net_ptp_poll != NULL) */
/*     sctk_net_ptp_poll (NULL); */

  tmp = &(sctk_ptp_list[arg->myself]);
  i = arg->communicator;
  communicator = &(tmp->communicators[i]);

  res = __sctk_perform_match_any_source_wait (communicator);
  __sctk_free_message_wait (communicator, tmp);

  if (arg->res != NULL)
    {
      if (communicator->wait_send_list != NULL)
	val++;
      if (communicator->wait_list != NULL)
	val++;
      sctk_nodebug ("%d message on %d on com %d (recv %d)", val,
		    arg->myself, i, res);

      *(arg->res) = val + res;
    }

}

static void
sctk_check_for_communicator_poll_one (sctk_message_wait_t * restrict arg)
{
  int i;
  sctk_ptp_data_t *restrict tmp;
  sctk_per_communicator_ptp_data_t *restrict communicator;

/*   if (sctk_net_ptp_poll != NULL) */
/*     sctk_net_ptp_poll (NULL); */

  tmp = &(sctk_ptp_list[arg->myself]);
  i = arg->communicator;
  communicator = &(tmp->communicators[i]);

  __sctk_perform_match_any_source_wait (communicator);
  __sctk_free_message_wait (communicator, tmp);

}

void
sctk_wait_message (sctk_request_t * msg)
{
  sctk_message_wait_t req;
  req.myself = msg->header.myself;
  req.communicator = msg->header.communicator;
  req.res = NULL;

  sctk_nodebug ("Wait message from %d to %d %p (%d)", msg->header.source,
		msg->header.destination, &(msg->completion_flag),
		msg->completion_flag);

  if (msg->completion_flag != 1)
    {
      sctk_check_for_communicator_poll_one (&req);

      if (msg->completion_flag != 1)
	{
	  sctk_thread_wait_for_value_and_poll
	    ((int *) &(msg->completion_flag), 1,
	     (void (*)(void *)) sctk_check_for_communicator_poll_one,
	     (void *) (&req));
	}
    }
  sctk_nodebug ("Wait message from %d to %d %p done", msg->header.source,
		msg->header.destination, &(msg->completion_flag));
}

void
sctk_perform_messages (const int task, const sctk_communicator_t com)
{
  sctk_message_wait_t req;
  int val = 0;
  req.myself = task;
  req.communicator = com;
  req.res = &val;
  sctk_check_for_communicator_poll (&req);
}

void
sctk_wait_all (const int task, const sctk_communicator_t com)
{
  sctk_message_wait_t req;
  sctk_ptp_data_t *tmp;
  sctk_per_communicator_ptp_data_t *communicator;
  int val = 0;
  req.myself = sctk_translate_to_global_rank (task, com);
  req.communicator = com;
  req.res = &val;

  tmp = &(sctk_ptp_list[task]);
  communicator = &(tmp->communicators[com]);
  if (communicator->wait_send_list != NULL)
    val++;
  if (communicator->wait_list != NULL)
    val++;
  sctk_nodebug ("Start %d message on %d %p\n", val, task, &val);

  sctk_check_for_communicator_poll (&req);
  sctk_thread_wait_for_value_and_poll (&val, 0, (void (*)(void *))
				       sctk_check_for_communicator_poll,
				       (void *) (&req));
  sctk_assert (communicator->wait_send_list == NULL);
  sctk_assert (communicator->wait_list == NULL);
  sctk_nodebug ("Start %d message on %d done %p\n", val, task, &val);
}
