/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL "http://www.cecill.info".                              # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */

typedef struct
{
  volatile int reinit;
  volatile int root_buffer;
  char buffer[collective_buffer_size];
} sctk_collective_msg_intern_msg_t;

typedef struct
{
  sctk_collective_msg_intern_msg_t msg;
  sctk_net_rdma_ack_t done;
} sctk_collective_msg_intern_t;

typedef sctk_collective_msg_intern_t
  sctk_collective_msg_t[collective_op_steps];

typedef struct
{
  sctk_collective_msg_t coll_recv[2];
  sctk_collective_msg_intern_t coll_send[2];
  int coll_nb;
} sctk_collective_t;

static sctk_collective_t *sctk_collective;

static void
sctk_net_preinit_collectives ()
{
  assume ((int) ceil (log2 (sctk_process_number)) < collective_op_steps);
  sctk_collective =
    sctk_iso_malloc (SCTK_MAX_COMMUNICATOR_NUMBER *
		     sizeof (sctk_collective_t));


}
static inline void
sctk_net_barrier_op_driver (sctk_collective_communications_t * com,
			    sctk_virtual_processor_t * my_vp)
{
  int step_number;
  int step_current;
  int acc;
  sctk_collective_msg_intern_t *coll_recv;
  sctk_collective_msg_intern_t *coll_send;
  int com_id;
  step_number = (int) ceil (log2 (com->involved_nb));
  step_current = 0;
  com_id = com->id;
  sctk_nodebug ("begin collective barrier %d",
		sctk_collective[com_id].coll_nb);

  coll_recv =
    (sctk_collective[com_id].coll_recv[sctk_collective[com_id].coll_nb % 2]);
  coll_send =
    &(sctk_collective[com_id].coll_send[sctk_collective[com_id].coll_nb % 2]);
  sctk_collective[com_id].coll_nb++;

  coll_send->msg.reinit = com->reinitialize;

  acc = 1;
  if (com->involved_nb == sctk_process_number)
    {
      for (step_current = 0; step_current < step_number; step_current++)
	{
	  int next;
	  next = (sctk_process_rank + acc) % sctk_process_number;

	  sctk_nodebug ("INIT for RDMA in collective %d -> %d",
			sctk_process_rank, next);
	  assume (sctk_net_rdma_write
		  (coll_send, &(coll_recv[step_current]),
		   sizeof (sctk_collective_msg_intern_msg_t), next,
		   &(coll_send->done), &(coll_recv[step_current].done)) == 0);
	  sctk_nodebug ("Wait for RDMA in collective step %d barrier %d %p",
			step_current, sctk_collective[com_id].coll_nb - 1,
			&(coll_send->done));
	  sctk_net_rdma_wait (&(coll_send->done));
	  sctk_nodebug
	    ("Wait for received RDMA in collective step %d barrier %d %p",
	     step_current, sctk_collective[com_id].coll_nb - 1,
	     &(coll_recv[step_current].done));
	  sctk_net_rdma_wait (&(coll_recv[step_current].done));

	  coll_send->msg.reinit =
	    coll_send->msg.reinit | coll_recv[step_current].msg.reinit;
	  acc *= 2;
	  sctk_nodebug ("End step %d barrier %d", step_current,
			sctk_collective[com_id].coll_nb - 1);
	}
    }
  else
    {
      not_implemented ();
    }

  com->reinitialize = coll_send->msg.reinit;
  memset (coll_recv, 0, sizeof (sctk_collective_msg_t));
  sctk_nodebug ("end collective barrier %d",
		sctk_collective[com_id].coll_nb - 1);
}

static inline void
sctk_net_broadcast_op_driver (sctk_collective_communications_t * com,
			      sctk_virtual_processor_t * my_vp,
			      const size_t elem_size,
			      const size_t nb_elem,
			      const sctk_datatype_t data_type,
			      void *data_in, void *data_out)
{
  int step_number;
  int step_current;
  int acc;
  sctk_collective_msg_intern_t *coll_recv;
  sctk_collective_msg_intern_t *coll_send;
  int com_id;
  int old_root_buffer = 0;
  sctk_nodebug ("begin collective broadcast");

  step_number = (int) ceil (log2 (com->involved_nb));
  step_current = 0;
  com_id = com->id;
  coll_recv =
    (sctk_collective[com_id].coll_recv[sctk_collective[com_id].coll_nb % 2]);
  coll_send =
    &(sctk_collective[com_id].coll_send[sctk_collective[com_id].coll_nb % 2]);
  sctk_collective[com_id].coll_nb++;

  coll_send->msg.reinit = com->reinitialize;
  coll_send->msg.root_buffer = 0;

  acc = 1;
  if (com->involved_nb == sctk_process_number)
    {
      if (elem_size * nb_elem <= collective_buffer_size)
	{
	  if (data_in)
	    {
	      coll_send->msg.root_buffer = 1;
	      old_root_buffer = 1;
	      memcpy (coll_send->msg.buffer, data_in, elem_size * nb_elem);
	    }

	  for (step_current = 0; step_current < step_number; step_current++)
	    {
	      int next;
	      next = (sctk_process_rank + acc) % sctk_process_number;

	      sctk_nodebug ("INIT for RDMA in collective %d -> %d",
			    sctk_process_rank, next);
	      assume (sctk_net_rdma_write
		      (coll_send, &(coll_recv[step_current]),
		       sizeof (sctk_collective_msg_intern_msg_t), next,
		       &(coll_send->done),
		       &(coll_recv[step_current].done)) == 0);
	      sctk_nodebug ("Wait for RDMA in collective");
	      sctk_net_rdma_wait (&(coll_send->done));
	      sctk_nodebug ("Wait for received RDMA in collective");
	      sctk_net_rdma_wait (&(coll_recv[step_current].done));

	      coll_send->msg.reinit =
		coll_send->msg.reinit | coll_recv[step_current].msg.reinit;
	      coll_send->msg.root_buffer =
		coll_send->msg.root_buffer | coll_recv[step_current].msg.
		root_buffer;

	      if (coll_send->msg.root_buffer && old_root_buffer == 0)
		{
		  sctk_nodebug ("Receive message");
		  memcpy (coll_send->msg.buffer,
			  coll_recv[step_current].msg.buffer,
			  elem_size * nb_elem);
		  old_root_buffer = 1;
		}

	      acc *= 2;
	      sctk_nodebug ("End step %d", step_current);
	    }

	  if (data_out)
	    {
	      memcpy (data_out, coll_send->msg.buffer, elem_size * nb_elem);
	    }
	}
      else
	{
	  not_implemented ();
	}
    }
  else
    {
      if (elem_size * nb_elem <= collective_buffer_size)
	{
	  not_implemented ();
	}
      else
	{
	  not_implemented ();
	}
    }
  com->reinitialize = coll_send->msg.reinit;
  memset (coll_recv, 0, sizeof (sctk_collective_msg_t));
  sctk_nodebug ("end collective broadcast");
}

static inline void
sctk_net_reduce_op_driver (sctk_collective_communications_t * com,
			   sctk_virtual_processor_t * my_vp,
			   const size_t elem_size,
			   const size_t nb_elem,
			   void (*func) (const void *, void *, size_t,
					 sctk_datatype_t),
			   const sctk_datatype_t data_type)
{
  int step_number;
  int step_current;
  int acc;
  sctk_collective_msg_intern_t *coll_recv;
  sctk_collective_msg_intern_t *coll_send;
  int com_id;
  sctk_nodebug ("begin collective reduce");

  step_number = (int) ceil (log2 (com->involved_nb));
  step_current = 0;
  com_id = com->id;
  coll_recv =
    (sctk_collective[com_id].coll_recv[sctk_collective[com_id].coll_nb % 2]);
  coll_send =
    &(sctk_collective[com_id].coll_send[sctk_collective[com_id].coll_nb % 2]);
  sctk_collective[com_id].coll_nb++;

  coll_send->msg.reinit = com->reinitialize;

  acc = 1;
  if (com->involved_nb == sctk_process_number)
    {
      int rank;
      int next;

      rank = sctk_process_rank;
      if (elem_size * nb_elem <= collective_buffer_size)
	{
	  memcpy (coll_send->msg.buffer, my_vp->data.data_in,
		  elem_size * nb_elem);
	  sctk_nodebug ("%d", *((int *) coll_send->msg.buffer));

	  /*descente */
	  for (step_current = 0; step_current < step_number; step_current++)
	    {
	      if (rank % (acc * 2) == 0)
		{
		  next = rank + acc;
		  if (next < sctk_process_number)
		    {
		      /*Recv from next */
		      sctk_net_rdma_wait (&(coll_recv[step_current].done));

		      sctk_nodebug ("recved %d",
				    *((int *) coll_recv[step_current].msg.
				      buffer));
		      func (coll_recv[step_current].msg.buffer,
			    coll_send->msg.buffer, nb_elem, data_type);
		      sctk_nodebug ("after func %d %p",
				    *((int *) coll_send->msg.buffer), func);

		      coll_send->msg.reinit =
			coll_send->msg.reinit | coll_recv[step_current].msg.
			reinit;
		    }
		}
	      else
		{
		  next = rank - acc;
		  /*send to next */

		  assume (sctk_net_rdma_write
			  (coll_send, &(coll_recv[step_current]),
			   sizeof (sctk_collective_msg_intern_msg_t), next,
			   &(coll_send->done),
			   &(coll_recv[step_current].done)) == 0);
		  sctk_net_rdma_wait (&(coll_send->done));
		  break;
		}


	      acc *= 2;
	      sctk_nodebug ("End step %d", step_current);
	    }

	  com->reinitialize = coll_send->msg.reinit;
	  memset (coll_recv, 0, sizeof (sctk_collective_msg_t));
	  if (rank == 0)
	    {
	      memcpy (my_vp->data.data_out, coll_send->msg.buffer,
		      elem_size * nb_elem);
	      sctk_nodebug ("Bcast send %d", *((int *) my_vp->data.data_out));
	      sctk_net_broadcast_op_driver (com, my_vp, elem_size, nb_elem,
					    data_type, my_vp->data.data_out,
					    NULL);
	    }
	  else
	    {
	      sctk_net_broadcast_op_driver (com, my_vp, elem_size, nb_elem,
					    data_type, NULL,
					    my_vp->data.data_out);
	      sctk_nodebug ("Bcast recv %d", *((int *) my_vp->data.data_out));
	    }
	}
      else
	{
	  not_implemented ();
	  com->reinitialize = coll_send->msg.reinit;
	  memset (coll_recv, 0, sizeof (sctk_collective_msg_t));
	}
    }
  else
    {
      not_implemented ();
      com->reinitialize = coll_send->msg.reinit;
      memset (coll_recv, 0, sizeof (sctk_collective_msg_t));
    }

  sctk_nodebug ("end collective reduce");
}

static void
sctk_net_collective_op_driver (sctk_collective_communications_t * com,
			       sctk_virtual_processor_t * my_vp,
			       const size_t elem_size,
			       const size_t nb_elem,
			       void (*func) (const void *, void *, size_t,
					     sctk_datatype_t),
			       const sctk_datatype_t data_type)
{
  sctk_nodebug ("begin collective");
  if (nb_elem == 0)
    {
      sctk_nodebug ("begin collective barrier");
      sctk_net_barrier_op_driver (com, my_vp);
      sctk_nodebug ("end collective barrier");
    }
  else
    {
      if (func == NULL)
	{
	  sctk_nodebug ("begin collective broadcast");
	  sctk_net_broadcast_op_driver (com, my_vp, elem_size, nb_elem,
					data_type, my_vp->data.data_in,
					my_vp->data.data_out);
	  sctk_nodebug ("end collective broadcast");
	}
      else
	{
	  sctk_nodebug ("begin collective reduce");
	  sctk_net_reduce_op_driver (com, my_vp, elem_size, nb_elem, func,
				     data_type);
	  sctk_nodebug ("end collective reduce");
	}
    }
  sctk_nodebug ("end collective");
}
