/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL "http://www.cecill.info".                              # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #   - CARRIBAULT Patrick patrick.carribault@cea.fr                     # */
/* #                                                                      # */
/* ######################################################################## */

#include <mpcthread.h>

#include <mpcmicrothread.h>

#include "mpcmicrothread_internal.h"
#include "sctk_alloc.h"
#include "sctk_context.h"
#include "sctk_debug.h"

static mpc_thread_once_t sctk_microthread_key_is_initialized =
  MPC_THREAD_ONCE_INIT;
  /* Key corresponding to the main micro-thread instance structure */
sctk_thread_key_t sctk_microthread_key;


  /* Function initializing a 'microthread_vp' structure */
static inline void sctk_microthread_init_microthread_vp_t
  (sctk_microthread_vp_t * self)
{
  /* Nothing to do */
  self->to_do_list = 0;
  self->to_do_list_next = 0;
  /* This vp is ready to accept microthreads */
  self->enable = 1;
}


/* Execute the operations of the VP 'self' */
static inline void
sctk_microthread_scheduler (sctk_microthread_vp_t * self)
{
  long i;
  long to_do;

  /* Grab the number of operations to perform */
  to_do = self->to_do_list;

  /* Iterate on the TO DO list and call the corresponding function */
  for (i = 0; i < to_do; i++)
    {
      sctk_nodebug ("sctk_microthread_scheduler: BEGIN func %d", i);

      /* Execute the i-th function */
      self->op_list[i].func ((void *) self->op_list[i].arg);

      sctk_nodebug ("sctk_microthread_scheduler: END func %d", i);
    }

  /* The list is now empty */
  self->to_do_list = 0;
}

/* Function called by the main scheduler when scheduling the corresponding VP
 */
static void *
sctk_microthread_slave_vp (void *arg)
{
  sctk_microthread_vp_t *self;

  sctk_nodebug ("sctk_microthread_slave_vp: START");

  /* Grab the information on the VP */
  self = (sctk_microthread_vp_t *) arg;

  /* While this VP is available */
  while (self->enable)
    {
      /* Is there anything to do? */
      if (self->to_do_list)
	{
	  /* Run the microthread scheduler */
	  sctk_microthread_scheduler (self);
	}
      else
	{
	  /* Fall back to the main scheduler */
	  sctk_thread_yield ();
	}
    }

  sctk_nodebug ("sctk_microthread_slave_vp: END");

  return NULL;
}

/* Add an operation to perform by the current VP */
int
sctk_microthread_add_task (void *(*func) (void *), void *arg,
			   long vp, long *step, sctk_microthread_t * task,
			   int val)
{
  sctk_microthread_vp_t *self;
  long i;

  sctk_nodebug ("sctk_microthread_add_task: VP=%d", vp);

  /* Grab the information on the currently scheduled VP */
  self = &(task->__list[vp]);

  /* Enqueue in the 'next' to-do list */
  i = self->to_do_list_next;
  self->op_list[i].func = MPC_MICROTHREAD_FUNC_T func;
  self->op_list[i].arg = arg;
  *step = i;

  i++;

  /*
     i -> If the number of op is known ('val' param), switch the 'to-do' to
     re-activate the thread execution
     -2 -> idem if this add task is the last one
   */
  if ((val == i) || (val == MPC_MICROTHREAD_LAST_TASK))
    {

      sctk_nodebug ("sctk_microthread_add_task: to_do switch forced");

      self->to_do_list_next = 0;
      self->to_do_list = i;
    }
  else
    {

      sctk_nodebug ("sctk_microthread_add_task: no to_do switch = %d", i);

      self->to_do_list_next = i;
    }

  return 0;
}

void *
sctk_microthread_first_task (void *arg)
{
  assume (arg != NULL);
  assume (sctk_thread_setspecific (sctk_microthread_key, arg) == 0);
  sctk_nodebug ("sctk_microthread_first_task: first_task on key %d",
		sctk_microthread_key);
  return NULL;
}

/* Initialization of the key named 'sctk_microthread_key' */
void
sctk_microthread_key_init ()
{
  sctk_thread_key_create (&sctk_microthread_key, NULL);
}

/* Initialization of a microthread instance */
int
sctk_microthread_init (long nb_vp, sctk_microthread_t * self)
{

  long i;
  int res;
  long nb_vp_t;

  /* Initialization of the sctk_microthread key (only once) */
  assume (mpc_thread_once (&sctk_microthread_key_is_initialized,
			   sctk_microthread_key_init) == 0);

  /* Force an even number of VPs because one loop during the execution is
   * unrolled by a factor of 2   */
  /* TODO maybe remove this computation is the corresponding loop is rolled back */
  nb_vp_t = nb_vp + nb_vp % 2;

  sctk_nodebug ("sctk_microthread_init: Ask for %d VP(s) (aligned to %d)",
		nb_vp, nb_vp_t);

  /* Memory allocation for all VPs */
  /* TODO maybe a function to 'create'/finalize a VP? */
  self->__list = (sctk_microthread_vp_t *) sctk_malloc (nb_vp_t * sizeof
							(sctk_microthread_vp_t));
  assume (self->__list != NULL);

  sctk_nodebug ("sctk_microthread_init: Ready to enter loop...");

  /* Update the number of VPs */
  self->__nb_vp = nb_vp;

  /* Create all VPs */
  for (i = 0; i < nb_vp_t; i++)
    {


      sctk_nodebug ("sctk_microthread_init: %d %d %p %ld", i,
		    sizeof (sctk_mctx_t), &((&(self->__list[i]))->vp_context),
		    ((long) &((&(self->__list[i]))->vp_context)) % 32);

      /* Get the set of registers */
      sctk_getcontext (&((&(self->__list[i]))->vp_context));
      /* Initialize the corresponding VP */
      sctk_microthread_init_microthread_vp_t ((&(self->__list[i])));

      sctk_nodebug ("sctk_microthread_init: %d done", i);
    }

  sctk_nodebug ("sctk_microthread_init: Ready to enter loop2...");

  /* Initialize slave VPs */
  for (i = 1; i < nb_vp; i++)
    {

      sctk_nodebug ("sctk_microthread_init: Before user thread create...");
      res =
	sctk_user_thread_create (&(self->__list[i].pid), NULL,
				 sctk_microthread_slave_vp,
				 &(self->__list[i]));
      assume (res == 0);

      sctk_nodebug ("sctk_microthread_init: After user thread create...");

      /* Force the thread creation by calling the main scheduler */
      sctk_thread_yield ();
    }

  /* Update the PID for the master VP */
  self->__list[0].pid = sctk_thread_self ();

  sctk_nodebug ("sctk_microthread_init: Ready to enter loop3...");

  /* Schedule a storage of the global structure 'sctk_microthread_t' */
  for (i = 1; i < nb_vp; i++)
    {
      long step;		/* Storing the current 'step' (index of current op) is not
				   necessary for this first execution */
      sctk_microthread_add_task (sctk_microthread_first_task, self, i, &step,
				 self, -1);
    }

  sctk_nodebug ("sctk_microthread_init: Ready to exec...");

  /* Launch the master execution of this micro thread instance */
  sctk_microthread_parallel_exec (self, MPC_MICROTHREAD_WAKE_VPS);

  return 0;
}




int
sctk_microthread_parallel_exec (sctk_microthread_t * task, int val)
{
  long i;
  sctk_microthread_t *old;	/* In case of nested microthread, store the old 'context' */
  sctk_microthread_vp_t *self;
  long running = 1;
  long __nb_vp;
  sctk_microthread_vp_t *__list;

  /* Store the old microthread context in case of nested microthreading */
  old = (sctk_microthread_t *) sctk_thread_getspecific (sctk_microthread_key);

  sctk_thread_setspecific (sctk_microthread_key, task);

  self = &(task->__list[0]);
  __nb_vp = task->__nb_vp;
  __list = task->__list;

  /* If needed, launch the other VPs by switching 'to_do' and 'to_do_next' */
  if (val == MPC_MICROTHREAD_WAKE_VPS)
    {

      sctk_nodebug ("Wake %d VPs", __nb_vp);

      for (i = 0; i < __nb_vp; i++)
	{
	  __list[i].to_do_list = __list[i].to_do_list_next;
	  __list[i].to_do_list_next = 0;
	}
    }

  /* Execute the operations of the current (master) VP */
  sctk_microthread_scheduler (self);

  /* Wait for everyone (implicit barrier) */
  while (running)
    {
      running = 0;
      /* TODO loop unrolled by a factor of 2: is it mandatory? */
      for (i = 1; i < __nb_vp; i += 2)
	{
	  running = running | (__list[i].to_do_list);
	  running = running | (__list[i + 1].to_do_list);
	}
      /* When someone is still working, a context swap is forced */
      if (running)
	{
	  sctk_thread_yield ();
	}
    }

  /* Restore the previous microthread context in case of nested microthreading */
  sctk_thread_setspecific (sctk_microthread_key, old);

  return 0;
}
