/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL "http://www.cecill.info".                              # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include "sctk_config.h"
#include "sctk_topologie.h"
#include "sctk_config.h"
#include "sctk_debug.h"
#include "sctk_thread.h"
#include "sctk_kernel_thread.h"

#ifdef SCTK_KERNEL_THREAD_USE_TLS
static __thread void *sctk_topologie_key;
#else
static kthread_key_t sctk_topologie_key;
#endif

#include <stdio.h>
#ifdef HAVE_UTSNAME
#include <sys/utsname.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <ctype.h>
#if defined(Linux_SYS)

#if defined(MPC_USE_LIBNUMA)
#include <numa.h>

#elif defined(MPC_USE_CPUSET)
#include <cpuset.h>

/* #else */
#include <sys/types.h>
#include <linux/unistd.h>
/* _syscall0 (pid_t, gettid) */
#define __USE_GNU
#include <sched.h>
#endif
#endif
#include <sched.h>
#if defined(HP_UX_SYS)
#include <sys/param.h>
#include <sys/pstat.h>
#endif
#define SCTK_MAX_PROCESSOR_NAME 100
#define SCTK_MAX_NODE_NAME 512
#define SCTK_MAX_CPUINFO_NUMBER 512
#define SCTK_LOCAL_VERSION_MAJOR 0
#define SCTK_LOCAL_VERSION_MINOR 1
typedef struct sctk_cpuinfo_s
{
  char model_name[SCTK_MAX_PROCESSOR_NAME + 1];
  float freqency;
  long cache_size;
  int physical_id;
  int numa_node;
  int cpu_id;
  int i;
} sctk_cpuinfo_t;

static int sctk_processor_number_on_node = 0;
static int sctk_node_number_on_node = 0;
static sctk_cpuinfo_t sctk_cpuinfos[SCTK_MAX_CPUINFO_NUMBER];
static char sctk_node_name[SCTK_MAX_NODE_NAME];
static long sctk_node_mem_free = 0;
static long sctk_node_mem_used = 0;
static long sctk_node_mem_total = 0;
static long sctk_node_swap_free = 0;
static long sctk_node_swap_used = 0;
static long sctk_node_swap_total = 0;


#if defined(Linux_SYS) && defined(MPC_USE_CPUSET)
static char *cpu_set_path = NULL;
#endif

static void
sctk_determine_processor_number ()
{
  long sctk_processor_number_on_node_tmp;
  /*   Détermination du nombre de processeurs disponibles */
#ifdef SunOS_SYS
  sctk_processor_number_on_node_tmp = sysconf (_SC_NPROCESSORS_CONF);
#elif defined(Linux_SYS)
  sctk_processor_number_on_node_tmp = sysconf (_SC_NPROCESSORS_CONF);
#elif defined(IRIX64_SYS)
  sctk_processor_number_on_node_tmp = sysconf (_SC_NPROC_CONF);
#elif defined(OSF1_SYS)
  sctk_processor_number_on_node_tmp = sysconf (_SC_NPROCESSORS_CONF);
#elif defined(AIX_SYS)
  sctk_processor_number_on_node_tmp = sysconf (_SC_NPROCESSORS_CONF);
#elif defined(HP_UX_SYS)
  do
    {
      struct pst_dynamic psd;
      pstat_getdynamic (&psd, sizeof (struct pst_dynamic), (size_t) 1, 0);
      sctk_processor_number_on_node_tmp = psd.psd_proc_cnt;
    }
  while (0);
#elif defined(WINDOWS_SYS)
  sctk_processor_number_on_node_tmp = pthread_num_processors_np ();
#else
#error unknown system
  sctk_processor_number_on_node = 1;
  sctk_abort ();
#endif
  sctk_processor_number_on_node = (int) sctk_processor_number_on_node_tmp;
  if (SCTK_MAX_CPUINFO_NUMBER <= sctk_processor_number_on_node)
    {
      fprintf (stderr, "Too many processors\n");
      sctk_abort ();
    }
}

/* static inline char* sctk_skip_token(char *p){ */
/*   while(isspace(*p)) p++; */
/*   while(*p && !isspace(*p)) p++; */
/*   while(isspace(*p)) p++; */
/*   return p; */
/* } */


int
sctk_get_cpu ()
{
  sctk_cpuinfo_t *old_cpu;
  old_cpu = (sctk_cpuinfo_t *) kthread_getspecific (sctk_topologie_key);

  if (old_cpu == NULL)
    {
      return -1;
    }
  else
    {
      return old_cpu->i;
    }
}

#if defined(MPC_USE_CPUSET)
static void
sctk_create_cpuset (int j)
{
  int res;
  char name[2048];
  cs_cpumask_t cmask;
  cs_memmask_t mmask;

  sprintf (name, "%s/MPC_proc_%02d", cpu_set_path, j);
  sctk_nodebug ("create %s", name);
  res = cs_create (name);
  assert (res == 0);

  /*Cpu init */
  res = cs_cpumask_clear (&cmask);
  assert (res == 0);
  res = cs_cpumask_add (&cmask, sctk_cpuinfos[j].cpu_id);
  assert (res == 0);
  sctk_nodebug ("add %d in %s", sctk_cpuinfos[j].cpu_id, name);
  res = cs_set_cpus (name, cmask);
  assert (res == 0);

  /*Mem init */
  res = cs_memmask_clear (&mmask);
  assert (res == 0);
  res = cs_memmask_add (&mmask, sctk_cpuinfos[j].numa_node);
  assert (res == 0);
  res = cs_set_mems (name, mmask);
  assert (res == 0);
}
#endif
int
sctk_bind_to_cpu (int i)
{
  sctk_cpuinfo_t *old_cpu;
  old_cpu = (sctk_cpuinfo_t *) kthread_getspecific (sctk_topologie_key);

  if (i >= 0)
    {
      kthread_setspecific (sctk_topologie_key, &(sctk_cpuinfos[i]));

#if defined(Linux_SYS)
#if defined(MPC_USE_LIBNUMA)
      {
	nodemask_t nodemask;

	nodemask = numa_no_nodes;
	nodemask_zero (&nodemask);
	nodemask_set (&nodemask, sctk_cpuinfos[i].numa_node);

	numa_bind (&nodemask);
      }
#elif defined(MPC_USE_CPUSET)
      {
	char name[2048];
	int res;
	static sctk_thread_mutex_t mutex = SCTK_THREAD_MUTEX_INITIALIZER;
	sprintf (name, "%s/MPC_proc_%02d", cpu_set_path, i);
	sctk_nodebug ("before %s asked %d (%s) %d/%d", cs_get_current (),
		      i, name, getpid (), getpid ());
	sctk_thread_mutex_lock (&mutex);
	cs_lock_libcpuset ();
	res = cs_add_task (name, getpid ());

	if (res != 0)
	  {
	    cs_unlock_libcpuset ();
	    sched_yield ();
	    cs_lock_libcpuset ();
	    res = cs_add_task (name, getpid ());
	  }
	if (res != 0)
	  {
	    cs_unlock_libcpuset ();
	    sched_yield ();
	    cs_lock_libcpuset ();
	    res = cs_add_task (name, getpid ());
	  }
	if (res != 0)
	  {
	    cs_unlock_libcpuset ();
	    sched_yield ();
	    cs_lock_libcpuset ();
	    res = cs_add_task (name, getpid ());
	  }


	assert (res == 0);
	cs_unlock_libcpuset ();
	sctk_thread_mutex_unlock (&mutex);
	sctk_nodebug ("after %s asked %d", cs_get_current (), i);
      }
#else
#if defined(CPU_ZERO)
      {
	cpu_set_t cpu;
	CPU_ZERO (&cpu);
	CPU_SET (i, &cpu);
	sched_setaffinity (0, CPU_SETSIZE, &cpu);
      }
#endif
#endif
#endif
    }

  if (old_cpu == NULL)
    {
      return -1;
    }
  else
    {
      return old_cpu->i;
    }
}


static void
sctk_update_numa ()
{
  int i;
  for (i = 0; i < SCTK_MAX_CPUINFO_NUMBER; i++)
    {
      sctk_cpuinfos[i].numa_node = 0;
    }
#if defined(Linux_SYS)
#if defined(MPC_USE_LIBNUMA)
  if (numa_available () >= 0)
    {
      int j;
      unsigned long *buffer;
      sctk_node_number_on_node = numa_max_node () + 1;
      buffer =
	(unsigned long *) malloc (sctk_processor_number_on_node *
				  sizeof (unsigned long));

      for (i = 0; i < sctk_node_number_on_node; i++)
	{
	  numa_node_to_cpus (i, buffer, sctk_processor_number_on_node);
	  for (j = 0; j < sctk_processor_number_on_node; j++)
	    {
	      if (buffer[j] < sctk_processor_number_on_node)
		sctk_cpuinfos[buffer[j]].numa_node = i;
	    }
	}
      sctk_free (buffer);
    }
  else
    {
      sctk_warning ("Unable to use libnuma\n");
    }
#elif defined(MPC_USE_CPUSET)
  if (strcmp (sctk_multithreading_mode, "ethread_mxn") == 0)
    {
      int j;
      int res;
      char name_mpc[2048];
      cs_cpumask_t rootmask;
      int cpu_set = 0;

      sprintf (name_mpc, "MPC_%d", getpid ());

      assert (cs_init () == 0);
      res = cpuset_simple_create (name_mpc, sctk_processor_number_on_node,
				  CS_AUTOCLEAN, NULL);
      assert (res == 0);

      res = cpuset_assign (name_mpc);
      assert (res == 0);
      sctk_node_number_on_node = cs_nr_nodes ();
      cpu_set_path = cs_get_current ();
      sctk_nodebug ("%d cpus %d nodes in %s",
		    sctk_processor_number_on_node,
		    sctk_node_number_on_node, cpu_set_path);

      i = 0;
      while (cpu_set != sctk_processor_number_on_node)
	{
	  cs_cpumask_t mask;
	  cs_get_node_cpus (i, &mask);
	  res = cs_get_cpus (cpu_set_path, &rootmask);
	  assert (res == 0);
	  sctk_nodebug ("Check node %d %d cpus in mask", i,
			cs_cpumask_cpus_count (&rootmask));
	  for (j = 0; j < sctk_processor_number_on_node; j++)
	    {
	      int first;
	      first = cs_cpumask_first (&rootmask);
	      if (cs_cpumask_cpu_present (&mask, first))
		{
		  sctk_cpuinfos[j].numa_node = i;
		  sctk_cpuinfos[j].cpu_id = first;
		  sctk_cpuinfos[j].i = j;
		  sctk_nodebug ("%d is on %02d", j, i);
		  cpu_set++;
		}
	      cs_cpumask_remove (&rootmask, first);
	    }
	  i++;
	}

      for (j = 0; j < sctk_processor_number_on_node; j++)
	{
	  cs_lock_libcpuset ();
	  sctk_create_cpuset (j);
	  cs_unlock_libcpuset ();

	}

    }
#else
  sctk_node_number_on_node = 0;
#endif
#endif
}

static void
sctk_update_memory (void)
{
}

#ifdef HAVE_UTSNAME
static struct utsname utsname;
#else
struct utsname
{
  char sysname[];
  char nodename[];
  char release[];
  char version[];
  char machine[];
};

static struct utsname utsname;

static int
uname (struct utsname *buf)
{
  buf->sysname = "";
  buf->nodename = "";
  buf->release = "";
  buf->version = "";
  buf->machine = "";
  return 0;
}

#endif

void
sctk_topologie_init ()
{
  int i;
  sctk_only_once ();
  sctk_print_version ("Topologie", SCTK_LOCAL_VERSION_MAJOR,
		      SCTK_LOCAL_VERSION_MINOR);

  kthread_key_create (&sctk_topologie_key, NULL);

  sctk_determine_processor_number ();
  memset (sctk_cpuinfos, 0,
	  sctk_processor_number_on_node * sizeof (sctk_cpuinfo_t));

#ifndef WINDOWS_SYS
  gethostname (sctk_node_name, SCTK_MAX_NODE_NAME);
#else
  sprintf (sctk_node_name, "localhost");
#endif

  uname (&utsname);

  for (i = 0; i < sctk_processor_number_on_node; i++)
    {
      sprintf (sctk_cpuinfos[i].model_name, "%s", utsname.machine);
    }
  sctk_update_memory ();
  sctk_update_numa ();
}

char *
sctk_get_node_name ()
{
  return sctk_node_name;
}

void
sctk_print_topologie (FILE * fd)
{
  int i;
  fprintf (fd, "Node %s: %s %s %s\n", sctk_node_name, utsname.sysname,
	   utsname.release, utsname.version);
  for (i = 0; i < sctk_processor_number_on_node; i++)
    {
      fprintf (fd, "\tProcessor %d (%d:%d):\n", i,
	       sctk_cpuinfos[i].physical_id, sctk_cpuinfos[i].numa_node);
      fprintf (fd, "\t\t* %s\n", sctk_cpuinfos[i].model_name);
      fprintf (fd, "\t\t* %f MHz\n", sctk_cpuinfos[i].freqency);
    }
  fprintf (fd, "\tMemory:\n");
  fprintf (fd, "\t\t* Total %ldk\n", sctk_node_mem_total);
  fprintf (fd, "\t\t* Used %ldk\n", sctk_node_mem_used);
  fprintf (fd, "\t\t* Free %ldk\n", sctk_node_mem_free);
  fprintf (fd, "\tSwap:\n");
  fprintf (fd, "\t\t* Total %ldk\n", sctk_node_swap_total);
  fprintf (fd, "\t\t* Used %ldk\n", sctk_node_swap_used);
  fprintf (fd, "\t\t* Free %ldk\n", sctk_node_swap_free);
  fprintf (fd, "\tNUMA: %d\n", sctk_is_numa_node ());
  if (sctk_is_numa_node ())
    {
      fprintf (fd, "\t\t* Node nb %d\n", sctk_node_number_on_node);
    }
}

char *
sctk_get_processor_name ()
{
  return sctk_cpuinfos[0].model_name;
}

int
sctk_get_node_from_cpu (int cpu)
{
  return sctk_cpuinfos[cpu].numa_node;
}

int
sctk_get_first_cpu_in_node (int node)
{
  int i;
  for (i = 0; i < sctk_processor_number_on_node; i++)
    {
      if (sctk_cpuinfos[i].numa_node == node)
	return i;
    }
  return 0;
}

int
sctk_get_cpu_number ()
{
  return sctk_processor_number_on_node;
}

int
sctk_set_cpu_number (int n)
{
  sctk_processor_number_on_node = n;
  return n;
}

int
sctk_is_numa_node ()
{
  return sctk_node_number_on_node != 0;
}
